# Copyright 2020-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Heptapod specific facilities for GitLab typed refs.

GitLab typed refs:

  GitLab typed refs are pointers to Mercurial changesets that the Rails
  application believes to be actual Git refs having significant
  types, such as branches or tags.

  Typed refs have short names beside the full ref path: a branch with name
  'foo' has full ref path 'refs/heads/foo', etc.

  GitLab typed refs are updated automatically if needed by each transaction,
  applying several rules tailored for
  good interaction with other GitLab components. They are not implemented
  as bookmarks because they are not meant to be exposed to Mercurial clients.

  GitLab typed refs are kept in files that represents the last state that was
  exposed to other GitLab components (notably, the Rails application). In case
  there is an ongoing transaction, the file represents the state at the
  beginning of the transaction, and is updated at the end of transaction.
  Its uses include:

  - sending pre- and post-receive calls to the Rails application based on
    comparisons with the previous state.
  - fast querying by GitLab branch in HGitaly. While it is possible for HGitaly
    to implement the mapping to GitLab branches in a stateless way, it is very
    bad for performance to do it in all read operations.

  The state cannot be reconstructed, since it records what other components
  have already seen. But it can be reinitialized: GitLab will fail to update
  the Merge Requests that should have been affected by the current transaction,
  fail to create new Pipelines and generally miss all the consequences of
  the current transaction. Initializing it at import or in an otherwise
  empty transaction should not create any problem.

  We currently track several types of refs, notably ``branches`` and ``tags``,
  each in its separate file for vastly different updating profiles: there is
  usually more tags than branches, but they don't get updated often. On the
  other hand, there are usually fewer visible branches, but each transaction
  updates at least one of them.

  File format: version preamble (common to all versions):

  - first 3 bytes: decimal string representation of the version number, padded
    with left zeros (spaces also accepted for reading)
  - byte 4: Unix End-Of-Line (LF, 0x10)

  File format: version 1 (after version preamble)

  - line-based, each line terminated the Unix way (LF, 0x10)
  - one GitLab ref per line
  - each line is made of the hexadecimal Mercurial changeset ID, followed by
    exactly one space character, then by the GitLab ref short name (bytes with
    no encoding assumption).

    Example for the branches file::

        01ca23fe01ca23fe01ca23fe01ca23fe01ca23fe branch/default

    Example for the tags file::

        01ca23fe01ca23fe01ca23fe01ca23fe01ca23fe v1.2.3


Default GitLab branch:
  This a GitLab branch, e.g., ``branch/default`` rather than a Mercurial
  branch, e.g., ``default``.

  It is stored inside the repository directory, but not in the main store.
  In principle, different shares could have different default GitLab branches,
  if that were to be useful.

  For comparison, GitLab uses the value of `HEAD` on Git repositories for
  this.
"""
GITLAB_TYPED_REFS_FILE_CURRENT_VERSION = 1

GITLAB_TYPED_REFS_MISSING = object()
"""Cacheable marker to avoid repeated read attempts if missing"""


def gitlab_typed_refs_file_name(type_name):
    return ('gitlab.' + type_name).encode()


def read_gitlab_typed_refs_file_version(fobj):
    """Read the version and leave `fobj` at the start of actual data.
    """
    version_line = fobj.read(4)
    # TODO proper exception
    assert version_line.endswith(b'\n')
    return int(version_line)  # ignores left 0 padding, whitespace and EOL


def write_gitlab_typed_refs_file_version(
        fobj,
        version=GITLAB_TYPED_REFS_FILE_CURRENT_VERSION):
    fobj.write(b"%03d\n" % version)


def read_gitlab_typed_refs(repo, type_name):
    """Return the GitLab named refs mapping for the given type, as a dict.

    The keys of the returned :class:`dict` are typed ref names (by opposition
    to full paths). Examples:

    - ``topic/stable/fix1`` (branch)
    - ``v1.2.3`` (tag).

    The values are Mercurial hexadecimal changeset ids, as :class:`bytes`
    """
    # TODO exceptions other than `FileNotFoundException`, which is fine
    with repo.svfs(gitlab_typed_refs_file_name(type_name), mode=b'rb') as fobj:
        assert read_gitlab_typed_refs_file_version(fobj) == 1
        return {ref_name.strip(): sha for sha, ref_name in
                (line.split(b' ', 1) for line in fobj)}


def remove_gitlab_typed_refs(repo, type_name):
    repo.svfs.unlink(gitlab_typed_refs_file_name(type_name))


def gitlab_typed_refs(repo, type_name):
    """Return GitLab named refs for the given type or GITLAB_TYPED_REFS_MISSING.

    The state file can still optional at this stage.
    Callers have to check for GITLAB_TYPED_REFS_MISSING and fall back to
    other means.

    Will eventually become a property on the repo object. Since our use case
    is different than what Mercurial usual filecaches are meant for, it
    is safer to do it directly in this first version.
    """
    # TODO rename glb as glr
    cache_attr = '_gitlab_refs_' + type_name
    glb = getattr(repo, cache_attr, None)
    if glb is not None:
        return glb

    try:
        glb = read_gitlab_typed_refs(repo, type_name)
    except FileNotFoundError:
        glb = GITLAB_TYPED_REFS_MISSING
    setattr(repo, cache_attr, glb)
    return glb


def write_gitlab_typed_refs(repo, type_name, refs_by_name):
    """Write a file for a type of GitLab refs.

    :param refs_by_name: :class:`dict` mapping typed ref names (example
      for ``type_name='branch'``: ``branch/stable``) to Mercurial hexadecimal
    changeset ids (as :class:`bytes`).
    """
    setattr(repo, '_gitlab_refs_' + type_name, None)
    with repo.lock():
        with repo.svfs(gitlab_typed_refs_file_name(type_name),
                       mode=b'wb',
                       atomictemp=True,
                       checkambig=True) as fobj:
            write_gitlab_typed_refs_file_version(fobj)
            for name, sha_hex in refs_by_name.items():
                fobj.write(b"%s %s\n" % (sha_hex, name))
            # atomictempfile does not expose flush(), so there's no
            # point trying an fsync(). Mercurial does not seem to use
            # fsync() anyway.



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
try:
    from pkg_resources import packaging
except ImportError:  # pragma no cover
    # one day, `packaging` might not be vendored in `pkg_resources` any more
    import packaging

import heptapod


def test_package_version():
    # the Version class accepts only PEP440 compliant version strings
    packaging.version.Version(heptapod.__version__)



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import os
from dulwich.tests.test_object_store import (
    DiskObjectStore,
    DiskObjectStoreTests,
)


class AlternatePathsTests(DiskObjectStoreTests):

    def test_read_alternate_paths(self):
        store = DiskObjectStore(self.store_dir)

        abs_path = os.path.abspath(os.path.normpath('/abspath'))
        # ensures in particular existence of the alternates file
        store.add_alternate_path(abs_path)
        self.assertEqual(set(store._read_alternate_paths()), {abs_path})

        store.add_alternate_path("relative-path")
        self.assertIn(os.path.join(store.path, "relative-path"),
                      set(store._read_alternate_paths()))

        # arguably, add_alternate_path() could strip comments.
        # Meanwhile it's more convenient to use it than to import INFODIR
        store.add_alternate_path("# comment")
        for alt_path in store._read_alternate_paths():
            self.assertNotIn("#", alt_path)



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from mercurial import cmdutil, commands, error


def mirror(ui, repo, *args, **kwargs):
    repo.ui.note(b"Heptapod mirror starting")
    gitlab_mirror = cmdutil.findcmd(b'gitlab-mirror', commands.table)[1][0]
    try:
        gitlab_mirror(ui, repo)
    except error.Abort:
        raise
    except Exception:
        repo.ui.traceback()
        raise
    repo.ui.note(b"Heptapod mirror done\n")
    return 0



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Interaction with GitLab for native Mercurial repos.

Although we don't need to convert native repos to Git, we still have to
maintain the GitLab branches, and fire appropriate notifications.
"""
from __future__ import absolute_import
import attr
import collections
import itertools

from mercurial.node import hex
from heptapod.gitlab.branch import (
    gitlab_branch_from_ref,
    gitlab_branch_ref,
)
from heptapod.gitlab.tag import gitlab_tag_ref

from .branch import (
    GITLAB_TYPED_REFS_MISSING,
    get_default_gitlab_branch,
    set_default_gitlab_branch,
    gitlab_branches,
    gitlab_tags,  # TODO move in separate module?
)
from .state_maintainer import GitLabStateMaintainer


@attr.s
class RefsByType:
    """Record GitLab refs, with a set for each type of ref.

    Types of ref can be `tags`, `refs` or whatever becomes useful to track.
    """
    heads = attr.ib(default=attr.Factory(set))
    tags = attr.ib(default=attr.Factory(set))

    def __iter__(self):
        return itertools.chain(self.heads, self.tags)

    def __bool__(self):
        return bool(self.heads) or bool(self.tags)


class NoGitStateMaintainer(GitLabStateMaintainer):
    """Maintain GitLab state for native Mercurial without any Git repository.

    Use of this class must be restricted to native Mercurial repositories
    running in "fully native" mode, but that is the concern of configuration,
    which should in turn be passed along by the Rails application or other
    relevant GitLab components.
    """

    def __init__(self, ui, repo):
        self.ui = ui
        self.repo = repo

        self.unfiltered_repo = self.repo.unfiltered()
        self._default_git_ref = None

        gl_branches = gitlab_branches(repo)
        if gl_branches is GITLAB_TYPED_REFS_MISSING:
            # should mean we are on an empty repo
            gl_branches = {}

        gl_tags = gitlab_tags(repo)
        if gl_tags is GITLAB_TYPED_REFS_MISSING:
            # should mean we are on an empty repo
            gl_tags = {}

        self.gitlab_refs = {gitlab_branch_ref(gl_branch): sha
                            for gl_branch, sha in gl_branches.items()}
        self.gitlab_refs.update((gitlab_tag_ref(gl_tag), sha)
                                for gl_tag, sha in gl_tags.items())

    def map_hg_get(self, sha):
        return sha

    def sync(self):
        """Main entry point for synchronisation of shared state with GitLab."""
        self.update_gitlab_references()

    def get_default_gitlab_ref(self):
        res = self._default_git_ref
        if res is None:
            gl_branch = get_default_gitlab_branch(self.repo)
            if gl_branch is not None:
                res = gitlab_branch_ref(gl_branch)
        return res

    def set_default_gitlab_ref(self, new_default_ref):
        new_gl_branch = gitlab_branch_from_ref(new_default_ref)
        self.repo.ui.note(
            b"Setting GitLab branch to %s" % new_gl_branch)
        set_default_gitlab_branch(self.repo, new_gl_branch)

        # cache invalidation
        self._default_git_ref = None

    def hg_sha_from_gitlab_sha(self, sha):
        return sha

    def gitlab_sha_from_hg_sha(self, sha):
        return sha

    def _filter_for_bookmarks(self, bms):
        # A simple compatibily with GitHandler, but truly, since we
        # don't use the branch_bookmark_suffix, we would be better off
        # not calling this method from the base class
        return [(bm, bm) for bm in bms]

    def extract_bookmarks(self, git_refs):
        """Update git_refs with current bookmarks."""
        for name, node in self.repo._bookmarks.items():
            git_refs[hex(node)].heads.add(gitlab_branch_ref(name))

    def extract_all_gitlab_refs(self):
        """Heptapod version of GitHandler.get_exportable() for native case.

        We should be able to do without tags, as they are handled separately
        (no persistent state in Mercurial repo, just notifications by
        :meth:heptapod_compare_tags)
        """
        gitlab_refs = collections.defaultdict(RefsByType)
        self.extract_bookmarks(gitlab_refs)
        self.extract_current_gitlab_branches(gitlab_refs)
        return gitlab_refs



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""A minimal py.test to make sure the testing harness works.
"""
import subprocess


def test_hg_install():
    vinfo = subprocess.check_output(('hg', 'version'))
    assert vinfo.startswith(b'Mercurial')



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Heptapod facilities for GitLab branches.

GitLab tags are a case of :mod:`typed refs <hgext3rd.heptapod.typed_ref>`,
we store them in their own state file, which represents the current view
by the other GitLab components, hence if there is an ongoing transaction,
it acts as a snapshot of the state at the beginning of the
transaction, so that we can notify them about the changes at the end
of the transaction.

Default GitLab branch:
  This a GitLab branch, e.g., ``branch/default`` rather than a Mercurial
  branch, e.g., ``default``.

  It is stored inside the repository directory, but not in the main store.
  In principle, different shares could have different default GitLab branches,
  if that were to be useful.

  For comparison, GitLab uses the value of `HEAD` on Git repositories for
  this.
"""
from mercurial.i18n import _
from mercurial import (
    hg,
    error,
)
from .typed_ref import (
    GITLAB_TYPED_REFS_MISSING,
    gitlab_typed_refs,
    read_gitlab_typed_refs,
    write_gitlab_typed_refs,
    remove_gitlab_typed_refs,
)
from .tag import *  # noqa: F401,F403 (reexports for retrocompat)

DEFAULT_GITLAB_BRANCH_FILE_NAME = b'default_gitlab_branch'

GITLAB_BRANCHES_MISSING = GITLAB_TYPED_REFS_MISSING  # retrocompat


def read_gitlab_branches(repo):
    return read_gitlab_typed_refs(repo, 'branches')


def gitlab_branches(repo):
    return gitlab_typed_refs(repo, 'branches')


def write_gitlab_branches(repo, gl_branches):
    """Write the GitLab branches mapping atomically.
    """
    return write_gitlab_typed_refs(repo, 'branches', gl_branches)


def remove_gitlab_branches(repo):
    remove_gitlab_typed_refs(repo, 'branches')


def get_default_gitlab_branch(repo):
    """Return the default GitLab branch name, or ``None`` if not set."""
    branch = repo.vfs.tryread(DEFAULT_GITLAB_BRANCH_FILE_NAME)
    # (hg 5.4) tryread returns empty strings for missing files
    if not branch:
        return None
    return branch


def set_default_gitlab_branch(repo, target):
    if not target:
        raise error.Abort(_("The default GitLab branch cannot be an "
                            "empty string."))
    shared_from = hg.sharedreposource(repo)
    if shared_from is not None:
        repo = shared_from

    with repo.wlock():
        repo.vfs.write(DEFAULT_GITLAB_BRANCH_FILE_NAME, target)



# Copyright 2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Interaction with Git repos.

It is not expected that the rise of the HGitaly project would remove
all interesting things to do with Git.
"""
from __future__ import absolute_import

try:
    from hgext3rd.hggit.git_handler import GitHandler
except ImportError:  # pragma: no cover (fallback for hg-git < 0.11)
    from hggit.git_handler import GitHandler

from heptapod.gitlab.branch import (
    gitlab_branch_from_ref as git_branch_from_ref,
)
from mercurial import (
    hg,
)
import os
import re

from .branch import (
    GITLAB_TYPED_REFS_MISSING,
    get_default_gitlab_branch,
    set_default_gitlab_branch,
    gitlab_branches,
    gitlab_tags,
    write_gitlab_branches,
    write_gitlab_tags,
)
from .special_ref import (
    special_refs,
    write_special_refs
)
from .keep_around import (
    iter_keep_arounds,
    init_keep_arounds,
)
from .state_maintainer import GitLabStateMaintainer


class HeptapodGitHandler(GitHandler, GitLabStateMaintainer):

    def __init__(self, *args, **kwargs):
        super(HeptapodGitHandler, self).__init__(*args, **kwargs)

        main_repo = hg.sharedreposource(self.repo)
        if main_repo is None:
            main_repo = self.repo

        self.gitdir = re.sub(br'\.hg$', b'', main_repo.root) + b'.git'
        self.unfiltered_repo = self.repo.unfiltered()
        self._default_git_ref = None

    @property
    def gitlab_refs(self):
        return self.git.refs

    def has_git_repo(self):
        return os.path.exists(self.gitdir)

    def get_default_gitlab_ref(self):
        # HEAD is the source of truth for GitLab default branch
        # For native projects, it's going to be the one stored
        # in Mercurial, returned by get_default_gitlab_branch(), and we are
        # maintaining consistency until we reach the HGitaly2 milestone
        # (i.e., as long as hg-git is used even for native projects)
        res = self._default_git_ref
        if res is None:
            res = self.git.refs.get_symrefs().get(b'HEAD')
            self._default_git_ref = res
        return res

    def ensure_gitlab_default_branch(self):
        """Init GitLab default branch file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab default branch :class:`bytes`.
        """
        repo = self.repo
        gl_branch = get_default_gitlab_branch(repo)
        if gl_branch is not None:
            return

        self.ui.note(b"Initializing GitLab default branch file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            gl_branch_ref = self.get_default_gitlab_ref()
            gl_branch = git_branch_from_ref(gl_branch_ref)
        else:
            gl_branch = b'master'

        set_default_gitlab_branch(repo, gl_branch)
        return gl_branch

    def set_default_gitlab_ref(self, new_default_ref):
        new_gl_branch = git_branch_from_ref(new_default_ref)
        self.repo.ui.note(
            b"Setting Git HEAD to %s and Hg default "
            b"GitLab branch to %s" % (new_default_ref, new_gl_branch))
        self.git.refs.set_symbolic_ref(b'HEAD', new_default_ref)
        set_default_gitlab_branch(self.repo, new_gl_branch)

        # cache invalidation
        self._default_git_ref = None

    hg_sha_from_gitlab_sha = GitHandler.map_hg_get
    gitlab_sha_from_hg_sha = GitHandler.map_git_get

    def extract_all_gitlab_refs(self):
        """Heptapod version of GitHandler.get_exportable().

        This rewraps :meth:`GitHandler.get_exportable` to add named branches
        and topics to the returned Git refs
        """
        git_refs = super(HeptapodGitHandler, self).get_exportable()
        self.extract_current_gitlab_branches(git_refs)
        return git_refs

    def export_commits(self):
        try:
            self.export_git_objects()
            self.update_gitlab_references()
        finally:
            self.save_map(self.map_file)

    def ensure_gitlab_branches(self):
        """Init GitLab branches state file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab branches :class:`dict`, same as
           ``gitlab_branches(self.repo)`` would.
        """
        repo = self.repo
        if gitlab_branches(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab branches state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            gl_branches = self.gitlab_branches_hg_shas(self.git.refs.as_dict())
        else:
            gl_branches = {}
        write_gitlab_branches(repo, gl_branches)
        return gl_branches

    def ensure_gitlab_tags(self):
        """Init GitLab tags state file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab tags :class:`dict`, same as
           ``gitlab_tags(self.repo)`` would.
        """
        repo = self.repo
        if gitlab_tags(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab tags state file "
                     b"for repo at '%s'" % repo.root)
        if self.has_git_repo():
            gl_tags = self.gitlab_tags_hg_shas(self.git.refs)
        else:
            gl_tags = {}
        write_gitlab_tags(repo, gl_tags)
        return gl_tags

    def ensure_gitlab_special_refs(self):
        """Init GitLab special refs file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab special refs :class:`dict`, same as
           ``special_refs(self.repo)`` would.
        """
        repo = self.repo
        if special_refs(repo) is not GITLAB_TYPED_REFS_MISSING:
            return

        self.ui.note(b"Initializing GitLab special refs state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            srefs = self.gitlab_special_refs_hg_shas(self.git.refs)
        else:
            srefs = {}
        write_special_refs(repo, srefs)
        return srefs

    def ensure_gitlab_keep_arounds(self):
        """Init GitLab keep arounds file from Git repository if needed.

        Nothing happens if the file is already present.

        :returns: ``None` if the file was already present. Otherwise,
           new GitLab keep-arounds :class:`set`.
        """
        repo = self.repo
        try:
            if next(iter_keep_arounds(repo)) is not GITLAB_TYPED_REFS_MISSING:
                return
        except StopIteration:
            return

        self.ui.note(b"Initializing GitLab keep-arounds state file "
                     b"for repo at '%s'" % repo.root)

        if self.has_git_repo():
            kas = self.gitlab_keep_arounds(self.git.refs)
        else:
            kas = ()
        init_keep_arounds(repo, kas)
        return kas



# Copyright 2019-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from pathlib import Path

from . import patch  # noqa: F401

__version__ = (  # noqa: F401
    Path(__file__).with_name('VERSION').read_text().strip())



# heptapod/hgweb.py - Heptapod HTTP interface for a directory of repositories.
#
# derived under GPL2+ from Mercurial's hgwebdir_mod.py, whose
# copyright holders are
#   Copyright 21 May 2005 - (c) 2005 Jake Edge <jake@edge2.net>
#   Copyright 2005, 2006 Matt Mackall <mpm@selenic.com>
#
# This file Copyright 2019 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

import logging
import gc
import os
import weakref

from mercurial.i18n import _

from mercurial.hgweb.common import (
    ErrorResponse,
    HTTP_SERVER_ERROR,
    HTTP_NOT_FOUND,
    cspvalues,
    statusmessage,
)

from mercurial import (
    error,
    extensions,
    hg,
    profiling,
    pycompat,
    ui as uimod,
)

from mercurial.hgweb import (
    hgweb_mod,
    request as requestmod,
)
from mercurial.repoview import _filteredrepotypes


logger = logging.getLogger(__name__)
# logging configuration will be initialized from the Mercurial global
# configuration, overriding this:
logging.basicConfig(level=logging.INFO)

ALLOWED_REMOTES = (
    '127.0.0.1',
    '::1',
    '',  # Unix Domain Socket
)


# Suffixes of HTTP headers that are directly forwarded in environment
FORWARDED_HEADERS_SUFFIXES = (
    'USERINFO_ID',
    'USERINFO_USERNAME',
    'USERINFO_NAME',
    'USERINFO_EMAIL',
    'PROJECT_PATH',
    'PROJECT_NAMESPACE_FULL_PATH',
)


def clear_repo_class(repo_class):
    _filteredrepotypes.pop(repo_class, None)


class HgServe(object):
    """WSGI application serving repositories under a given root path

    The repositories are expected in the `heptapod.repositories-root`
    Mercurial configuration.

    This works under full trust of the incoming request: callers are either
    `gitlab-rails` or `gitlab-workhorse`.
    """
    def __init__(self, conf_path=None, baseui=None):
        """On python3, conf_path is str."""
        self.baseui = baseui
        if baseui:
            self.ui = baseui.copy()
        else:
            self.ui = uimod.ui.load()
        self.motd = None
        if not baseui:
            # set up environment for new ui
            extensions.loadall(self.ui)
            extensions.populateui(self.ui)
        if conf_path is not None:
            for i, rcf in enumerate(conf_path.split(':')):
                # ui.readconfig does a blank open(), but will later
                # store the file path (needing bytes, of course)
                rcf = pycompat.fsencode(rcf)
                if i == 0 and not os.path.exists(rcf):
                    raise error.Abort(_(b'config file %s not found!') % rcf)
                logger.info("Loading configuration from %r", rcf)
                self.ui.readconfig(rcf, trust=True)

        root = self.ui.config(b'heptapod', b'repositories-root')
        if root is None:
            raise ValueError("heptapod.repositories-root is not configured.")
        self.repos_root = root

    def apply_heptapod_headers(self, environ):
        perm_user = environ.get('HTTP_X_HEPTAPOD_PERMISSION_USER')
        if perm_user is not None and environ['REMOTE_ADDR'] in ALLOWED_REMOTES:
            environ['REMOTE_USER'] = perm_user
        for hepta_key in FORWARDED_HEADERS_SUFFIXES:
            hepta_val = environ.pop('HTTP_X_HEPTAPOD_' + hepta_key, None)
            if hepta_val is not None:
                environ['HEPTAPOD_' + hepta_key] = hepta_val
        # special case for GL_REPOSITORY, forwarded as such to match
        # system environment variable name in subprocess calls
        gl_repo = environ.pop('HTTP_X_HEPTAPOD_GL_REPOSITORY', None)
        if gl_repo is not None:
            environ['GL_REPOSITORY'] = gl_repo
            split_repo = gl_repo.split('-', 1)
            if len(split_repo) == 2:
                environ['HEPTAPOD_REPOSITORY_USAGE'] = split_repo[0]
                environ['HEPTAPOD_PROJECT_ID'] = split_repo[1]

    def __call__(self, env, respond):
        self.apply_heptapod_headers(env)
        profile = self.ui.configbool(b'profiling', b'enabled')
        with profiling.profile(self.ui, enabled=profile):
            try:
                for r in self._runwsgi(env, respond):
                    yield r
            finally:
                # There are known cycles in localrepository that prevent
                # those objects (and tons of held references) from being
                # collected through normal refcounting. We mitigate those
                # leaks by performing an explicit GC on every request.
                # TODO remove this once leaks are fixed.
                # TODO only run this on requests that create localrepository
                # instances instead of every request.
                gc.collect()

    def load_repo(self, uri_path, env):
        repo_path = os.path.join(self.repos_root, uri_path)
        if not os.path.isdir(os.path.join(repo_path, b'.hg')):
            # hg.repository() would raise a RepoError which is
            # not qualified enough to distinguish it cleanly (just
            # the message)
            raise ErrorResponse(HTTP_NOT_FOUND, "Not Found")
        logger.info("loading repo at %r", repo_path)
        repo = hg.repository(self.ui, repo_path)
        weakref.finalize(repo, clear_repo_class, repo.unfiltered().__class__)

        # setting native mode booleans
        native = env.get('HTTP_X_HEPTAPOD_HG_NATIVE')
        if native:
            repo.ui.setconfig(b'heptapod', b'native',
                              pycompat.sysbytes(native))
        if repo.ui.configbool(b'heptapod', b'native'):
            no_git = env.get('HTTP_X_HEPTAPOD_NO_GIT')
            if no_git is not None:
                repo.ui.setconfig(b'heptapod', b'no-git',
                                  pycompat.sysbytes(no_git))
        return repo

    def _runwsgi(self, env, respond):
        baseurl = self.ui.config(b'web', b'baseurl')
        # TODO simplify without bringing too much code duplication
        first_parse = requestmod.parserequestfromenv(env, altbaseurl=baseurl)
        uri_path = first_parse.dispatchpath.strip(b'/')
        try:
            # Actual parsing of WSGI environment to take into account our
            # repository path component.
            req = requestmod.parserequestfromenv(
                env, reponame=uri_path,
                altbaseurl=baseurl,
                # Reuse wrapped body file object otherwise state
                # tracking can get confused.
                bodyfh=first_parse.bodyfh)
            res = requestmod.wsgiresponse(req, respond)
            csp = cspvalues(self.ui)[0]
            if csp:
                res.headers[b'Content-Security-Policy'] = csp

            try:
                repo = self.load_repo(uri_path, env)
                return hgweb_mod.hgweb(repo).run_wsgi(req, res)
            except IOError as inst:
                raise ErrorResponse(HTTP_SERVER_ERROR, inst.strerror)
            except error.RepoError as inst:  # pragma: no cover
                # This case can't happen because hgweb_mod catches them
                # already. To be removed in Heptapod 0.8
                raise ErrorResponse(HTTP_SERVER_ERROR, bytes(inst))

        except ErrorResponse as e:
            # not passing the 'message' kwarg here gives us the standard
            # human readable version of the code (e.g., '404 Not Found').
            # Details about the error are provided in the body.
            res.status = statusmessage(e.code)
            res.headers[b'Content-Type'] = b'text/plain; charset=UTF-8'
            # ErrorResponse.message is produced by pycompat.sysstr
            # Python2: the encoding could be anything, and it's not
            #          not even clear that `encoding.encoding` would apply.
            #          At worst, our Content-Type is wrong.
            # Python3: it could be emitted originally as `str` or have been
            #          decoded to `latin-1` by `sysstr()`. In the latter
            #          case, the upstream usage of `sysstr()` would be plain
            #          wrong.
            # That was in theory. In practice, all messages generated
            # by hgweb are actually ASCII as of Mercurial 5.4,
            # except maybe repetition of user requested file.
            # pycompat.sysbytes() encodes to UTF-8 in py3
            res.setbodybytes(pycompat.sysbytes(e.message) + b'\n')
            return res.sendresponse()

    @classmethod
    def default_app(cls):
        HEPTAPOD_HGRC = os.environ.get('HEPTAPOD_HGRC')
        if HEPTAPOD_HGRC:
            return HgServe(conf_path=HEPTAPOD_HGRC)


hgserve = HgServe.default_app()



# Copyright 2020-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
"""Integration tests for the GitLab state maintainer with no Git repo.

In this case, "integration" means that we're testing the main methods,
and the whole Hook logic. We may also need to create
complex transactions, with several commits, rebases, phase changes etc.
"""
from __future__ import absolute_import

import contextlib
import pytest
import re

from heptapod.gitlab import hooks
from heptapod.gitlab.prune_reasons import (
    HeadPruneReason,
    BookmarkRemoved,
    BranchClosed,
    TopicPublished,
    WildHeadResolved,
)
from mercurial_testhelpers.repo_wrapper import NULL_ID
from mercurial_testhelpers.util import as_bytes
from heptapod.testhelpers import (
    RepoWrapper,
)
from heptapod.testhelpers.gitlab import (
    patch_gitlab_hooks,
    GitLabStateMaintainerFixture,
)
from heptapod.hooks.tests.utils import switch_user
from hgext3rd.heptapod.branch import (
    set_default_gitlab_branch,
    gitlab_branches,
    read_gitlab_branches,
    write_gitlab_branches,
)
from mercurial import (
    encoding,
    error,
    pycompat,
    scmutil,
    ui as uimod,
)
from heptapod.gitlab.change import (
    GitLabRefChange as GitRefChange,
    ZERO_SHA,
)
from ...no_git import (
    NoGitStateMaintainer,
)
from ..utils import common_config, HEPTAPOD_REQUIRED_HGRC


parametrize = pytest.mark.parametrize


# not covering a log interceptor for tests is not a break of coverage
# perhaps this utility should move to pytest-mercurial
def patch_ui_warnings(monkeypatch, records):  # pragma no cover

    def warn(self, *a):
        records.append(a)

    monkeypatch.setattr(uimod.ui, 'warn', warn)


@contextlib.contextmanager
def gitlab_state_fixture_gen(tmpdir, monkeypatch):
    """Common fixture generator."""
    with GitLabStateMaintainerFixture.init(
            tmpdir, monkeypatch,
            hg_config=common_config()) as fixture:
        fixture.hg_repo_wrapper.set_config('heptapod.native', True)
        fixture.hg_repo_wrapper.set_config('heptapod.no-git', True)
        yield fixture


@pytest.fixture()
def empty_fixture(tmpdir, monkeypatch):
    """Minimal fixture where the repository is empty."""
    with gitlab_state_fixture_gen(tmpdir, monkeypatch) as fixture:
        yield fixture


@pytest.fixture()
def main_fixture(tmpdir, monkeypatch):
    """A fixture with two public changesets in the Mercurial repository.

    The changesets are kept on the fixture object as additional attributes
    ``base_ctx`` and ``ctx1``.

    It is up to the test using this fixture to share state with GitLab or not.
    """
    with gitlab_state_fixture_gen(tmpdir, monkeypatch) as fixture:
        wrapper = fixture.hg_repo_wrapper
        fixture.base_ctx = wrapper.commit_file('foo',
                                               content='foo0',
                                               message='default0')
        fixture.ctx1 = wrapper.commit_file('foo',
                                           content='foo1',
                                           message="default1")
        wrapper.set_phase('public', ['.'])
        yield fixture


def set_allow_bookmarks(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.bookmarks-on-named-branches', value)


def set_allow_multiple_heads(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(b'heptapod', b'allow-multiple-heads', value)


def set_prune_closed_branches(repo_wrapper, value):
    repo_wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-newly-closed-branches', value)


def test_basic(main_fixture):
    repo = main_fixture.hg_repo_wrapper
    notifs = main_fixture.gitlab_notifs

    repo.command('gitlab-mirror')

    sha = main_fixture.ctx1.hex()
    assert notifs == [
        ('pre-receive', ({}, {b'refs/heads/branch/default': (ZERO_SHA, sha)})),
        ('post-receive', ({},
                          {b'refs/heads/branch/default': (ZERO_SHA, sha)})),
    ]

    assert read_gitlab_branches(repo.repo) == {
        b'branch/default': repo.repo[b'.'].hex()
    }


@parametrize('pushvar', ('skip-ci', 'mr-create'))
def test_pushvars(main_fixture, pushvar):
    wrapper = main_fixture.hg_repo_wrapper
    notifs = main_fixture.gitlab_notifs

    if pushvar == 'skip-ci':
        hookargs = {b'USERVAR_CI.SKIP': b''}
        push_opts = {'ci.skip': ''}
    elif pushvar == 'mr-create':
        hookargs = {b'USERVAR_MERGE_REQUEST.CREATE': b''}
        push_opts = {'merge_request.create': ''}
    # no default case, better have an error instead

    with wrapper.repo.transaction(b'tests') as txn:
        txn.hookargs = hookargs
        wrapper.command('gitlab-mirror')

    sha = main_fixture.ctx1.hex()
    assert notifs == [
        ('pre-receive', ({}, {b'refs/heads/branch/default': (ZERO_SHA, sha)})),
        ('post-receive', ({},
                          {b'refs/heads/branch/default': (ZERO_SHA, sha)},
                          push_opts)),
    ]


def test_alternate_encoding(main_fixture, monkeypatch):
    """Round trip with branch name not in UTF-8.

    It's not clear whether this is valid on the GitLab side, but it should
    be transparent in here: what comes in comes out.
    """
    fixture = main_fixture
    notifs = fixture.gitlab_notifs
    wrapper = fixture.hg_repo_wrapper

    wrapper.command('gitlab-mirror')
    fixture.clear_gitlab_notifs()

    # Even though it's been initialized from the `HGENCODING` environment
    # variable, the encoding is a global.
    monkeypatch.setattr(encoding, 'encoding', b'latin-1')

    ctx = wrapper.write_commit('encoded', branch=b'pr\xe9parations')
    wrapper.command('gitlab-mirror')

    # not using existing helpers to avoid making the test tautological
    gitlab_branch = b'branch/pr\xe9parations'
    gitlab_ref = b'refs/heads/' + gitlab_branch
    sha = ctx.hex()

    # Both in GitLab branches state file and in notifs,
    # we have the exact same bytes as provided
    assert notifs == [
        ('pre-receive', ({}, {gitlab_ref: (ZERO_SHA, sha)})),
        ('post-receive', ({}, {gitlab_ref: (ZERO_SHA, sha)})),
    ]
    assert read_gitlab_branches(wrapper.repo)[gitlab_branch] == sha


def test_wiki(main_fixture):
    fixture = main_fixture
    notifs = fixture.gitlab_notifs
    wrapper = fixture.hg_repo_wrapper

    wrapper.repo.ui.environ[b'GL_REPOSITORY'] = b'wiki-251'
    wrapper.command('gitlab-mirror')

    sha = fixture.ctx1.hex()
    # for wikis, we duplicate `branch/default` as `master`
    changes = {}, {b'refs/heads/branch/default': (ZERO_SHA, sha),
                   b'refs/heads/master': (ZERO_SHA, sha),
                   }
    assert notifs == [
        ('pre-receive', changes),
        ('post-receive', changes),
    ]


def test_tags_simple(main_fixture):
    repo = main_fixture.hg_repo_wrapper
    hg_repo = repo.repo
    notifs = main_fixture.gitlab_notifs
    base_ctx = main_fixture.base_ctx

    # Creation
    repo.command('tag', b'v1.2.3', rev=base_ctx.hex())
    repo.command('gitlab-mirror')

    branches = gitlab_branches(hg_repo)
    assert list(branches) == [b'branch/default']
    first_default_branch_sha = branches[b'branch/default']

    tagged_sha_0 = base_ctx.hex()

    changes = {}, {
        b'refs/heads/branch/default': (ZERO_SHA, first_default_branch_sha),
        b'refs/tags/v1.2.3': (ZERO_SHA, tagged_sha_0),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()

    # Modification
    repo.command('tag', b'v1.2.3', rev=b'1', force=True)
    repo.command('gitlab-mirror')

    tagged_sha_1 = hg_repo[1].hex()

    new_default_branch_sha = hg_repo[b'tip'].hex()
    changes = {}, {
        b'refs/heads/branch/default': (first_default_branch_sha,
                                       new_default_branch_sha),
        b'refs/tags/v1.2.3': (tagged_sha_0, tagged_sha_1),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()

    repo.command('tag', b'v1.2.3', remove=True)
    repo.command('gitlab-mirror')

    last_default_branch_sha = hg_repo[b'tip'].hex()
    changes = {}, {
        b'refs/heads/branch/default': (new_default_branch_sha,
                                       last_default_branch_sha),
        b'refs/tags/v1.2.3': (tagged_sha_1, ZERO_SHA),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]


def test_tags_permission(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    hg_repo = wrapper.repo
    base_ctx = fixture.base_ctx

    fixture.activate_mirror()  # we'll need the mirror in-transaction

    wrapper.set_config('web', 'allow-push', 'writer')
    wrapper.set_config('web', 'allow-publish', 'publisher')

    # Publisher can create tag
    switch_user(wrapper, 'publisher')
    with wrapper.tags_sanitizer():
        wrapper.command('tag', b'v1.2.3', rev=base_ctx.hex())
        assert set(hg_repo.tags()) == {b'v1.2.3', b'tip'}

    # With simple write permission, every change in tags is refused
    switch_user(wrapper, 'writer')
    with wrapper.tags_sanitizer(), pytest.raises(error.Abort) as exc_info:
        wrapper.command('tag', b'v6.5.4')
    exc = exc_info.value
    assert b'permission' in exc.args[0]
    assert b'6.5.4' in exc.hint

    with wrapper.tags_sanitizer(), pytest.raises(error.Abort) as exc_info:
        wrapper.command('tag', b'v1.2.3', force=True)
    exc = exc_info.value
    assert b'permission' in exc.args[0]
    assert b'1.2.3' in exc.hint

    with wrapper.tags_sanitizer(), pytest.raises(error.Abort) as exc_info:
        wrapper.command('tag', b'v1.2.3', remove=True)
    exc = exc_info.value
    assert b'permission' in exc.args[0]
    assert b'1.2.3' in exc.hint

    # Publisher can change and remove tags
    switch_user(wrapper, 'publisher')
    with wrapper.tags_sanitizer():
        wrapper.command('tag', b'v1.2.3', force=True)

    with wrapper.tags_sanitizer():
        wrapper.command('tag', b'v1.2.3', remove=True)

    assert set(hg_repo.tags()) == {b'tip'}


def test_tags_escaping_removal(main_fixture):
    wrapper = main_fixture.hg_repo_wrapper
    hg_repo = wrapper.repo
    notifs = main_fixture.gitlab_notifs

    # a tag with a space character
    hg_tag = b'Version 1.2.3'
    gitlab_tag_ref = b'refs/tags/Version_1.2.3'
    wrapper.command('tag', hg_tag)
    wrapper.command('gitlab-mirror')

    assert set(hg_repo.tags()) == {hg_tag, b'tip'}
    tagged_sha = main_fixture.ctx1.hex()

    branches = gitlab_branches(hg_repo)
    assert list(branches) == [b'branch/default']

    branch_sha1 = branches[b'branch/default']
    changes = {}, {
        b'refs/heads/branch/default': (ZERO_SHA, branch_sha1),
        gitlab_tag_ref: (ZERO_SHA, tagged_sha),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()

    # Pushing something else does not count as a tag modification.
    # This asserts that heptapod#464 is fixed.
    branch_sha2 = wrapper.commit_file('something-else').hex()
    wrapper.command('gitlab-mirror')

    changes = {}, {
        b'refs/heads/branch/default': (branch_sha1, branch_sha2),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()

    # Change is still possible
    wrapper.command('tag', hg_tag, force=True)
    wrapper.command('gitlab-mirror')
    assert set(hg_repo.tags()) == {hg_tag, b'tip'}
    branch_sha3 = gitlab_branches(hg_repo)[b'branch/default']
    changes = {}, {
        b'refs/heads/branch/default': (branch_sha2, branch_sha3),
        gitlab_tag_ref: (tagged_sha, branch_sha2),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()

    # Removal is still possible
    wrapper.command('tag', hg_tag, remove=True)
    wrapper.command('gitlab-mirror')
    assert set(hg_repo.tags()) == {b'tip'}

    branch_sha4 = gitlab_branches(hg_repo)[b'branch/default']
    changes = {}, {
        b'refs/heads/branch/default': (branch_sha3, branch_sha4),
        gitlab_tag_ref: (branch_sha2, ZERO_SHA),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]
    main_fixture.clear_gitlab_notifs()


def test_tags_obsolete(empty_fixture):
    fixture = empty_fixture

    # we'll need to perform a pull in order to amend a tagged changeset
    # and rebase the tagging changeset in a single transaction.
    src_path = fixture.base_path / 'src.hg'
    src = RepoWrapper.init(src_path, config=common_config())

    dest = fixture.hg_repo_wrapper
    fixture.activate_mirror()

    def dest_pull():
        dest.command('pull', as_bytes(src_path), force=True)

    # Creation
    src.commit_file('foo')

    # no problem on receiving side with obsolescence of changesets
    # that it never received
    src.commit_file('bar')
    src.command("amend", message=b'amend before exchange')
    dest_pull()

    src.command('tag', b'v1.2.3')
    tag_ctx = scmutil.revsingle(src.repo, b'.')
    dest_pull()

    tagged = scmutil.revsingle(src.repo, b'v1.2.3')
    src.update_bin(tagged.node())
    # let's create a topic from the changeset to be amended,
    # so that it will still be visible after amendment.
    src.commit_file('foo', topic='stacked')

    # amending the tagged changeset
    src.update_bin(tagged.node())
    src_path.join("foo", "amending")
    src.command('amend', message=b'amend')
    src.command('rebase', rev=[tag_ctx.hex()])
    rebased_tag_ctx = scmutil.revsingle(src.repo, b'tip')

    with pytest.raises(error.Abort) as exc_info:
        dest_pull()
    assert re.search(br'tag.*v1\.2\.3.*obsolete', exc_info.value.args[0])

    src.prune(rebased_tag_ctx.hex())
    dest_pull()


def test_tip_obsolete(empty_fixture):
    """Since 'tip' is a tag, we don't to refuse making it obsolete."""
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper
    wrapper.commit_file('foo')

    fixture.activate_mirror()  # we'll need the mirror in-transaction

    ctx = wrapper.commit_file('foo')
    wrapper.prune(ctx.hex())  # no error

    # yet the tip of the unfiltered repo is obsolete
    # (if this changes in a future Mercurial version, we may get rid of
    # this assertion)
    assert b'tip' in wrapper.repo.unfiltered()[ctx.node()].tags()


def test_share(main_fixture):
    main_wrapper = main_fixture.hg_repo_wrapper
    base_ctx = main_fixture.base_ctx
    tmpdir = main_fixture.base_path

    # let's start with some commits known to GitLab
    main_wrapper.command('gitlab-mirror')

    # TODO make a helper on fixture for that
    assert set(gitlab_branches(main_wrapper.repo)) == {b'branch/default'}

    # now let's make a share
    dest_wrapper = main_wrapper.share(tmpdir.join('share.hg'))
    dest_wrapper.write_commit('bar', message='other0',
                              branch='other', parent=base_ctx.node())
    dest_wrapper.command('gitlab-mirror')
    # of course, since the write went through another localrepo instance,
    # we need to bypass the cache
    assert set(read_gitlab_branches(main_wrapper.repo)) == {b'branch/default',
                                                            b'branch/other'}


def test_bookmarks(empty_fixture):
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper

    # this activates in particular the gitlab-mirror hook
    wrapper.repo.ui.readconfig(HEPTAPOD_REQUIRED_HGRC, trust=True)

    set_allow_bookmarks(wrapper, True)

    base = wrapper.commit_file('foo', message="unbookmarked")
    bk1 = wrapper.commit_file('foo', message="book1")
    bk2 = wrapper.commit_file('foo', message="book2")
    wrapper.command('bookmark', b'zebook1', rev=bk1.hex())
    wrapper.command('bookmark', b'zebook2', rev=bk2.hex())
    assert read_gitlab_branches(wrapper.repo) == {b'branch/default': bk2.hex(),
                                                  b'zebook1': bk1.hex(),
                                                  b'zebook2': bk2.hex()}
    ctx = wrapper.commit_file('foo', message="new default head", parent=base)
    assert read_gitlab_branches(wrapper.repo) == {b'branch/default': ctx.hex(),
                                                  b'zebook1': bk1.hex(),
                                                  b'zebook2': bk2.hex()}

    # and multiple heads are still refused
    with pytest.raises(error.Abort) as exc:
        wrapper.commit_file('foo', message="2nd default head", parent=base)
    assert b'multiple heads' in exc.value.args[0]
    hint_lead, hint_shas = exc.value.hint.split(b': ')
    assert hint_lead == b'2 heads'
    assert ctx.hex()[:12] in set(hint_shas.split())


def test_bookmarks_prune(main_fixture):
    fixture = main_fixture
    server = fixture.hg_repo_wrapper
    base_ctx = fixture.base_ctx
    notifs = main_fixture.gitlab_notifs

    repo = server.repo
    default1_ctx = fixture.ctx1
    server.command('bookmark', b'zebook', rev=base_ctx.hex())
    # not being in a transaction accepts the bookmark immediately
    server.command('gitlab-mirror')

    fixture.clear_gitlab_notifs()

    assert gitlab_branches(repo) == {
        b'branch/default': default1_ctx.hex(),
        b'zebook': base_ctx.hex(),
    }

    fixture.activate_mirror()
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(base_ctx.node()) == []
    assert read_gitlab_branches(repo) == {
        b'branch/default': default1_ctx.hex(),
    }

    changes = {b'zebook': BookmarkRemoved(base_ctx.hex())}, {
        b'refs/heads/zebook': (base_ctx.hex(), ZERO_SHA)
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]


def test_bookmarks_mask_branch_prune(main_fixture):
    fixture = main_fixture
    server = fixture.hg_repo_wrapper
    notifs = main_fixture.gitlab_notifs
    repo = server.repo
    default_sha = fixture.ctx1.hex()

    server.command('gitlab-mirror')
    # just checking our assumptions
    assert gitlab_branches(repo) == {b'branch/default': default_sha}

    # we need to test the branch masking on a branch
    # that's not the GitLab default (which is protected)
    server.write_commit('foo', branch='other', message='other1')

    fixture.activate_mirror()
    set_allow_bookmarks(server, True)
    head = server.repo[b'tip']
    server.command('bookmark', b'zebook', rev=head.hex())
    head_sha = head.hex()
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        b'zebook': head_sha,
    }

    fixture.clear_gitlab_notifs()
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(head.node()) == []
    assert read_gitlab_branches(repo) == {
        b'branch/default': default_sha,
        b'branch/other': head_sha,
    }

    changes = {b'zebook': BookmarkRemoved(head_sha)}, {
        b'refs/heads/zebook': (head_sha, ZERO_SHA),
        b'refs/heads/branch/other': (ZERO_SHA, head_sha),
    }
    assert notifs == [('pre-receive', changes), ('post-receive', changes)]


def test_bookmarks_obsolete(empty_fixture):
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo

    set_allow_bookmarks(wrapper, True)

    draft1 = wrapper.commit_file('foo', message='Initial draft')
    wrapper.command('bookmark', b'zebook', rev=draft1.hex(), inactive=True)
    wrapper.command('gitlab-mirror')

    # just checking our assumptions.
    assert gitlab_branches(repo) == {b'zebook': draft1.hex()}

    # let's get in-transaction
    fixture.activate_mirror()

    draft2 = wrapper.commit_file('foo', message='amended', parent=NULL_ID)
    with pytest.raises(error.Abort) as exc_info:
        wrapper.prune(draft1.hex(), successors=[draft2.hex()])
    assert re.search(br'bookmark.*zebook.*obsolete', exc_info.value.args[0])


def test_bookmarks_dont_mask_default_branch(main_fixture):
    fixture = main_fixture
    server = fixture.hg_repo_wrapper
    repo = server.repo
    notifs = fixture.gitlab_notifs
    sha = fixture.ctx1.hex()

    server.command('gitlab-mirror')
    # just checking our assumptions
    assert gitlab_branches(repo) == {b'branch/default': sha}

    fixture.activate_mirror()
    set_allow_bookmarks(server, True)
    head = server.repo[b'tip']

    fixture.clear_gitlab_notifs()
    server.command('bookmark', b'zebook', rev=head.hex())

    # the default branch is not pruned
    assert gitlab_branches(repo) == {
        b'branch/default': sha,
        b'zebook': sha,
    }

    changes = {b'refs/heads/zebook': (ZERO_SHA, sha)}
    assert notifs == [('pre-receive', ({}, changes)),
                      ('post-receive', ({}, changes)),
                      ]

    fixture.clear_gitlab_notifs()
    server.command('bookmark', b'zebook', delete=True)

    assert server.repo.nodebookmarks(head.node()) == []
    assert gitlab_branches(repo) == {b'branch/default': sha}
    changes = {b'refs/heads/zebook': (sha, ZERO_SHA)}
    prunes = {b'zebook': BookmarkRemoved(sha)}

    assert notifs == [('pre-receive', (prunes, changes)),
                      ('post-receive', (prunes, changes)),
                      ]


@parametrize('branch_name', ('default', 'other'))
def test_change_gitlab_default_branch(empty_fixture, branch_name):
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper

    wrapper.write_commit('foo', message="other0",
                         branch=branch_name,
                         topic='initial')
    wrapper.command('gitlab-mirror')

    fixture.assert_default_gitlab_branch('topic/%s/initial' % branch_name)

    wrapper.set_phase('public', ['.'])
    wrapper.command('gitlab-mirror')

    fixture.assert_default_gitlab_branch('branch/' + branch_name)


def test_change_gitlab_default_branch_nothing_new(empty_fixture):
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper

    wrapper.write_commit('foo', message="other0",
                         branch='default',
                         topic='initial')
    wrapper.command('gitlab-mirror')

    fixture.assert_default_gitlab_branch('topic/default/initial')

    wrapper.write_commit('foo', message="same gitlab branch")
    wrapper.command('gitlab-mirror')

    fixture.assert_default_gitlab_branch('topic/default/initial')


def test_closed_branch(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs
    base_ctx = fixture.base_ctx
    default_sha = fixture.ctx1.hex()

    other_ctx = wrapper.write_commit('foo', message="other0",
                                     branch='other',
                                     parent=base_ctx)
    wrapper.command('gitlab-mirror')

    to_close_sha = other_ctx.hex()
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        b'branch/other': to_close_sha,
    }
    set_prune_closed_branches(wrapper, True)
    fixture.activate_mirror()
    fixture.clear_gitlab_notifs()

    wrapper.command('commit', message=b"closing other",
                    close_branch=True)
    # we'll let GitLab do the pruning so that it can use the closing
    # sha for Merge Request detection.
    assert gitlab_branches(repo) == {b'branch/default': default_sha}

    closing_sha = scmutil.revsingle(repo, b'.').hex()
    prune_reason = {b'branch/other':
                    BranchClosed([(closing_sha, [to_close_sha])])}
    changes = {b'refs/heads/branch/other': (to_close_sha, ZERO_SHA)}

    assert notifs == [('pre-receive', (prune_reason, changes)),
                      ('post-receive', (prune_reason, changes))]


def test_previously_closed_branch_not_pruned(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    base_ctx = fixture.base_ctx
    default_sha = fixture.ctx1.hex()

    wrapper.commit_file('foo', message="other0",
                        branch='other',
                        parent=base_ctx)
    # let's prepare a closed branch that hasn't been pruned
    set_prune_closed_branches(wrapper, False)
    wrapper.command('commit', message=b"closing other", close_branch=True)
    closing_sha = scmutil.revsingle(repo, b'.').hex()
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {b'branch/default': default_sha,
                                     b'branch/other': closing_sha}

    # subsequent calls won't prune it...
    set_prune_closed_branches(wrapper, True)
    wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-previously-closed-branches', False)
    wrapper.command('gitlab-mirror')

    assert gitlab_branches(repo).get(b'branch/other') == closing_sha

    # until we flip the right switch
    wrapper.repo.ui.setconfig(
        b'experimental', b'hg-git.prune-previously-closed-branches', True)
    wrapper.command('gitlab-mirror')

    assert b'branch/other' not in gitlab_branches(repo)


def test_closed_branch_unknown_to_gitlab(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo

    wrapper.commit_file('foo', message="other0",
                        branch='other',
                        parent=fixture.base_ctx)

    # the `other` branch being mirrored as already closed, will trigger
    # a prune request that should be ignored in order not to fail
    wrapper.command('commit', message=b"closing other", close_branch=True)
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {b'branch/default': fixture.ctx1.hex()}


def test_closed_default_branch(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs
    default_sha = fixture.ctx1.hex()

    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {b'branch/default': default_sha}
    set_default_gitlab_branch(repo, b'branch/default')

    set_prune_closed_branches(wrapper, True)
    wrapper.command('commit', message=b"closing default!", close_branch=True)

    # On native repos, this is a user error

    fixture.clear_gitlab_notifs()
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    assert re.search(br'prune.*default branch', exc_info.value.args[0])

    # and nothing happened
    assert not notifs
    assert gitlab_branches(repo) == {b'branch/default': default_sha}


def test_multiple_heads_merge(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs

    set_allow_multiple_heads(wrapper, True)
    first = wrapper.repo[b'tip']
    second = wrapper.commit_file('bar', message="second head",
                                 branch='default',
                                 parent=fixture.base_ctx)
    first_sha = first.hex()
    second_sha = second.hex()
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {
        b'wild/' + first_sha: first_sha,
        b'wild/' + second_sha: second_sha,
        # the most recently added revision always wins
        b'branch/default': second_sha,
    }
    wrapper.command('merge')
    wrapper.command('commit', message=b'merging heads')
    merge_sha = wrapper.repo[b'.'].hex()

    fixture.clear_gitlab_notifs()
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {
        b'branch/default': merge_sha,
    }

    prune_reasons = {
        b'wild/' + first_sha: WildHeadResolved(first_sha),
        b'wild/' + second_sha: WildHeadResolved(second_sha),
    }
    changes = {
        b'refs/heads/wild/' + first_sha: (first_sha, ZERO_SHA),
        b'refs/heads/wild/' + second_sha: (second_sha, ZERO_SHA),
        b'refs/heads/branch/default': (second_sha, merge_sha)
    }

    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_push_multiple_heads_switch_branch(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    base_ctx = fixture.base_ctx

    set_allow_multiple_heads(wrapper, True)
    # second head on default branch (easy to merge for later test)
    default_head2 = wrapper.commit_file(
        'bar', message="default head 2", parent=base_ctx)
    default_heads_shas = {fixture.ctx1.hex(), default_head2.hex()}

    wrapper.command('gitlab-mirror')

    gl_branches = gitlab_branches(repo)
    assert set(gl_branches.values()) == default_heads_shas
    assert len(gl_branches) == 3

    other_head = wrapper.commit_file(
        'foo', message="other", branch='other')
    other_sha = other_head.hex()
    wrapper.command('gitlab-mirror')
    gl_branches = gitlab_branches(repo)
    assert set(gl_branches.values()) == default_heads_shas | {other_sha, }
    assert len(gl_branches) == 4

    # now let's add a topic on top of one of those wild 'default' heads
    topic_ctx = wrapper.commit_file(
        'foo', message="on topic",
        topic='zetop',
        parent=base_ctx)

    wrapper.command('gitlab-mirror')
    gl_branches = gitlab_branches(repo)
    assert set(b for b in gl_branches if not b.startswith(b'wild/')) == {
        b'branch/default', b'branch/other', b'topic/default/zetop'}

    assert set(sha
               for name, sha in gl_branches.items()
               if name.startswith(b'wild/')) == default_heads_shas
    assert gl_branches[b'branch/default'] in default_heads_shas
    assert gl_branches[b'branch/other'] == other_sha
    assert gl_branches[b'topic/default/zetop'] == topic_ctx.hex()


def test_push_multiple_heads_refuse(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    base_ctx = fixture.base_ctx

    # base line with GitLab branches state file already initialized
    wrapper.command('gitlab-mirror')
    default_sha = fixture.ctx1.hex()

    wrapper.commit_file('bar', message="default head 2", parent=base_ctx)

    # default behaviour:
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    # with explicit config:
    set_allow_multiple_heads(wrapper, False)
    with pytest.raises(error.Abort) as exc_info:
        wrapper.command('gitlab-mirror')

    assert b'multiple heads' in exc_info.value.args[0]
    assert gitlab_branches(repo) == {b'branch/default': default_sha}


def test_topic_pruned(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs
    default_sha = fixture.ctx1.hex()

    topic_ctx = wrapper.commit_file(
        'foo', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    fixture.clear_gitlab_notifs()

    topic_sha = topic_ctx.hex()
    topic_gl_branch = b'topic/default/zzetop'
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        topic_gl_branch: topic_sha}

    wrapper.prune(b'zzetop')
    wrapper.command('gitlab-mirror')

    topic_ref = b'refs/heads/topic/default/zzetop'
    topic_change = {topic_ref: (topic_sha, ZERO_SHA)}
    prune_reason = {topic_gl_branch: HeadPruneReason()}

    assert gitlab_branches(repo) == {b'branch/default': default_sha}
    assert notifs == [('pre-receive', (prune_reason, topic_change)),
                      ('post-receive', (prune_reason, topic_change))]


def test_topic_amended(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs
    default_sha = fixture.ctx1.hex()

    topic_ctx = wrapper.commit_file(
        'foo', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    fixture.clear_gitlab_notifs()

    topic_sha = topic_ctx.hex()
    topic_gl_branch = b'topic/default/zzetop'
    assert gitlab_branches(repo) == {b'branch/default': default_sha,
                                     topic_gl_branch: topic_sha}

    amended_ctx = wrapper.amend_file(
        'foo', message=b'amend1')
    wrapper.command('gitlab-mirror')

    amended_sha = amended_ctx.hex()
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        topic_gl_branch: amended_sha}

    topic_change = {
        b'refs/heads/topic/default/zzetop': (topic_sha, amended_sha)
    }

    assert notifs == [('pre-receive', ({}, topic_change)),
                      ('post-receive', ({}, topic_change))]


def test_topic_ff_publish(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs
    default_sha = fixture.ctx1.hex()

    topic_ctx = wrapper.commit_file(
        'foo', message='in ff topic', topic='zzetop')
    wrapper.command('gitlab-mirror')
    fixture.clear_gitlab_notifs()

    topic_gl_branch = b'topic/default/zzetop'
    topic_before_sha = topic_ctx.hex()
    assert gitlab_branches(repo) == {b'branch/default': default_sha,
                                     topic_gl_branch: topic_before_sha}

    wrapper.set_phase('public', ['zzetop'])
    wrapper.command('gitlab-mirror')

    assert gitlab_branches(repo) == {b'branch/default': topic_before_sha}

    topic_ref = b'refs/heads/topic/default/zzetop'

    prune_reasons = {topic_gl_branch: TopicPublished(topic_before_sha)}
    changes = {
        b'refs/heads/branch/default': (default_sha, topic_before_sha),
        topic_ref: (topic_before_sha, ZERO_SHA)
    }

    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


@parametrize('before_publish', ('add-rebase', 'rebase'))
def test_topic_publish(main_fixture, before_publish):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs

    topic_ctx = wrapper.commit_file(
        'bar', message='in topic',
        parent=fixture.base_ctx,
        topic='zzetop')
    wrapper.command('gitlab-mirror')
    fixture.clear_gitlab_notifs()
    topic_gl_branch = b'topic/default/zzetop'

    default_before_sha = fixture.ctx1.hex()
    topic_before_sha = topic_ctx.hex()
    assert gitlab_branches(repo) == {
        b'branch/default': default_before_sha,
        topic_gl_branch: topic_before_sha
    }

    if before_publish == 'add-rebase':
        wrapper.commit_file('zz', message='topic addition')

    wrapper.command('rebase', rev=[b'topic(zzetop)'])
    rebased_sha = scmutil.revsingle(repo, b'zzetop').hex()
    wrapper.set_phase('public', ['zzetop'])
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {b'branch/default': rebased_sha}

    prune_reasons = {topic_gl_branch: TopicPublished(rebased_sha)}
    changes = {
        b'refs/heads/branch/default': (default_before_sha, rebased_sha),
        b'refs/heads/topic/default/zzetop': (topic_before_sha, ZERO_SHA),
    }
    assert notifs == [('pre-receive', (prune_reasons, changes)),
                      ('post-receive', (prune_reasons, changes))]


def test_topic_git_escape(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper

    # Only test names that are possible with topics in the first place
    invalid_git_ref_names = [
        ".starts-with-dot",
        "..",
        "ends-with.lock",
        "ends-with-dot.",
    ]
    for name in invalid_git_ref_names:
        wrapper.write_commit('zz', message='in topic', topic=name)

    # This fails if topic names that are invalid in Git are not escaped
    wrapper.command('gitlab-mirror')


def test_topic_clear_publish(main_fixture):
    """The topic head seen from Git is public and has changed topic.

    This is the test for heptapod#265
    The starting point can be considered to be corrupted: any topic change
    should have updated the GitLab branch for the topic. In this scenario
    the change is a removal, wich should have pruned the Git branch, but
    somehow the GitLab branch got updated to the new changeset, that doesn't
    have a topic.
    """
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    notifs = fixture.gitlab_notifs

    topic_before_ctx = wrapper.commit_file(
        'zz', message='in topic', topic='zzetop')
    wrapper.command('gitlab-mirror')

    topic_gl_branch = b'topic/default/zzetop'

    default_before_sha = fixture.ctx1.hex()
    topic_before_sha = topic_before_ctx.hex()
    gitlab_branches_before = gitlab_branches(repo)
    assert gitlab_branches_before == {b'branch/default': default_before_sha,
                                      topic_gl_branch: topic_before_sha}

    wrapper.command('topics', rev=[b'.'], clear=True)
    topic_after_sha = scmutil.revsingle(repo, b'.').hex()
    wrapper.command('gitlab-mirror')
    wrapper.set_phase('public', ['.'])
    forced_topic_before_sha = default_before_sha

    # here's the main point:
    corrupted_gitlab_branches = gitlab_branches(repo)
    corrupted_gitlab_branches[topic_gl_branch] = forced_topic_before_sha
    write_gitlab_branches(repo, corrupted_gitlab_branches)

    fixture.clear_gitlab_notifs()
    # This used to raise LookupError
    wrapper.command('gitlab-mirror')

    assert gitlab_branches(repo) == {b'branch/default': topic_after_sha}
    changes = (
        {topic_gl_branch: TopicPublished(forced_topic_before_sha)},
        {b'refs/heads/topic/default/zzetop': (forced_topic_before_sha,
                                              ZERO_SHA)},
    )
    assert notifs == [('pre-receive', changes),
                      ('post-receive', changes)]


def test_topic_branch_change(main_fixture):
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper
    repo = wrapper.repo
    base_ctx = fixture.base_ctx
    default_sha = fixture.ctx1.hex()

    other_ctx = wrapper.commit_file(
        'bar', message='other0', parent=base_ctx, branch='other')
    topic_other_ctx = wrapper.commit_file(
        'zz', message='in topic', topic='zzetop')

    wrapper.command('gitlab-mirror')

    other_sha, topic_other_sha = other_ctx.hex(), topic_other_ctx.hex()
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        b'branch/other': other_sha,
        b'topic/other/zzetop': topic_other_sha,
    }

    wrapper.command('rebase', rev=[b'zzetop'], dest=b'default')
    wrapper.command('gitlab-mirror')
    assert gitlab_branches(repo) == {
        b'branch/default': default_sha,
        b'branch/other': other_sha,
        b'topic/default/zzetop': scmutil.revsingle(repo, b'zzetop').hex(),
    }


# git.test_integration has test_analyse_vanished_topic_unknown_to_git, that
# does not make sense for native repos, since it is about a changeset present
# in Mercurial and not in the Git auxiliary repository,
# instead of about branches/refs being out of sync


def test_heptapod_notify_gitlab(tmpdir, monkeypatch):
    wrapper = RepoWrapper.init(tmpdir.join('repo'))
    handler = NoGitStateMaintainer(wrapper.repo.ui, wrapper.repo)

    notifs = []

    def trigger_failures(name, changes):
        if b'exc' in changes[1]:
            raise RuntimeError('trigger_failures: ' + name)
        if b'code_one' in changes[1]:
            return 1, b'', 'hook refusal'
        return 0, b'Consider a MR for ' + name.encode(), ''

    patch_gitlab_hooks(monkeypatch, notifs, action=trigger_failures)

    # minimal valid change
    change = GitRefChange(b'master', b'before', b'after')
    hook = hooks.PreReceive(wrapper.repo)
    handler.heptapod_notify_gitlab(hook, [], {b'master': change},
                                   allow_error=False)
    assert notifs == [('pre-receive',
                       ([], {b'master': (b'before', b'after')})
                       )]
    del notifs[:]

    with pytest.raises(error.Abort) as exc_info:
        handler.heptapod_notify_gitlab(hook, [], {b'code_one': change},
                                       allow_error=False)
    assert b"hook refusal" in exc_info.value.args[0]

    with pytest.raises(RuntimeError) as exc_info:
        handler.heptapod_notify_gitlab(hook, [], {b'exc': change},
                                       allow_error=False)
    assert "trigger_failures" in exc_info.value.args[0]

    # case where exception is triggered yet is accepted
    errors = []

    def record_ui_error(*args):
        errors.append(args)
    monkeypatch.setattr(handler.repo.ui, 'error', record_ui_error)
    handler.heptapod_notify_gitlab(hook, [], {b'exc': change},
                                   allow_error=True)
    assert errors[0][0].splitlines()[:2] == [
        (b"GitLab update error (GitLab 'pre-receive' hook): "
         + pycompat.sysbytes(
             # this repr() differs between py2 and py3
             repr(RuntimeError('trigger_failures: pre-receive')))),
        b"Traceback (most recent call last):"
    ]


# git.test_integration has a test_heptapod_notify_gitlab_native
# that obviously doesn't make sense for NoGitStateMaintainer, which is always
# native.


def test_subrepos(main_fixture):
    """See heptapod#310 and heptapod#311."""
    fixture = main_fixture
    wrapper = fixture.hg_repo_wrapper

    # TODO wrapper.path when available
    main_path = fixture.base_path / 'repo.hg'
    nested_path = main_path / 'nested'
    nested = RepoWrapper.init(nested_path)
    nested.write_commit("bar", content="in nested")
    main_path.join('.hgsub').write("\n".join((
        "nested = ../bar",
        "[subpaths]",  # reproduction of heptapod#310
        "foo2=bar2",
    )))

    wrapper.command(b'add', subrepos=True)
    wrapper.command(b'commit', subrepos=True, message=b"invalid")
    wrapper.command('gitlab-mirror')

    wrapper.repo.ui.setconfig(b'hooks', b'precommit',
                              b'python:heptapod.hooks.subrepos.forbid_commit')
    with pytest.raises(error.Abort) as exc_info:
        wrapper.write_commit("foo")
    assert re.search(b'cannot commit.*there are subrepos',
                     exc_info.value.args[0])

    # Now let's simulate update from NULL revision, as Heptapod's
    # use of shares would do
    wrapper.command(b'update', b'0000000000000000000000', clean=True)
    nested_path.remove()

    with pytest.raises(error.Abort) as exc_info:
        wrapper.command(b'update', b'default')
    assert b'not supported' in exc_info.value.args[0]


def test_default_gitlab_ref(empty_fixture):
    fixture = empty_fixture
    wrapper = fixture.hg_repo_wrapper

    handler = NoGitStateMaintainer(wrapper.repo.ui, wrapper.repo)
    # default Git value
    assert handler.get_default_gitlab_ref() is None
    other_ref = b'refs/heads/something'
    handler.set_default_gitlab_ref(other_ref)

    # instance level cache got invalidated
    assert handler.get_default_gitlab_ref() == other_ref


# git.test_integration has tests for the various `ensure_*` methods,
# starting from there.
# The `NoGitStateMaintainer` class does not implement them



# Copyright 2019-2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from .. import util


def test_format_shas():
    fmt = util.format_shas
    shas = [b'dbad978fb2c3694e0ff8f6ac6889ffafe34ca0b6',
            b'88670e10840db7ccfec5cb0c590970907ac3b382',
            b'393dc94367af44ec57e7cd3a804ddc77e998e564',
            b'55e51a162f6fd259fab1c81395c7ba4d469335c2',
            b'bfd903562a133d228dd91c66954706b1cba35756',
            b'02ad4cac53815a65b890a090ef344268d149a4bb']

    assert fmt(shas[:1]) == b'dbad978fb2c3'
    assert fmt(shas[:2]) == b'dbad978fb2c3 88670e10840d'
    assert fmt(shas) == (
        b'dbad978fb2c3 88670e10840d 393dc94367af 55e51a162f6f and 2 others'
    )
    assert fmt(shas, limit=2) == b'dbad978fb2c3 88670e10840d and 4 others'

    # works with sets as well as lists (see heptapod#441)
    assert fmt(set(shas)).endswith(b'and 2 others')



from pathlib import Path
from setuptools import setup, find_packages

with open('install-requirements.txt', 'r') as install_reqf:
    install_req = [req.strip() for req in install_reqf]

VERSION_FILE = 'VERSION'

setup(
    name='heptapod',
    version=Path('heptapod', VERSION_FILE).read_text().strip(),
    author='Georges Racinet',
    author_email='georges.racinet@octobus.net',
    url='https://foss.heptapod.net/heptapod/py-heptapod',
    description="Heptapod server-side Mercurial hooks, extension, etc.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='hg mercurial heptapod gitlab',
    license='GPLv2+',
    package_data=dict(heptapod=['*.hgrc', VERSION_FILE]),
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved"
        " :: GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 2 :: Only",
        "Topic :: Software Development :: Version Control :: Mercurial",
    ],
    install_requires=install_req,
)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
import subprocess
from ..git import GitRepo


def test_branches(tmpdir):
    server = tmpdir / 'server.git'
    # server_repo is bare, hence we cannot commit directly in there
    server_repo = GitRepo.init(server)

    client = tmpdir / 'client'
    subprocess.call(('git', 'init', str(client)))
    subprocess.call(('git', 'remote', 'add', 'origin', str(server)),
                    cwd=str(client))
    subprocess.call(('git', 'config', 'user.name', 'John Doe'),
                    cwd=str(client))
    subprocess.call(('git', 'config', 'user.email', 'jdoe@heptapod.example'),
                    cwd=str(client))

    (client / 'foo').write('foo')
    subprocess.call(('git', 'add', 'foo'), cwd=str(client))

    subprocess.call(('git', 'commit', '-m', 'init commit', '--no-gpg-sign'),
                    cwd=str(client))
    subprocess.call(('git', 'push', 'origin', 'master'), cwd=str(client))

    assert server_repo.branch_titles() == {b'master': b'init commit'}

    # correct commit SHA is read from server repo
    sha = server_repo.get_branch_sha('master')
    client_log = subprocess.check_output(['git', 'log', '--oneline', sha],
                                         cwd=str(client))
    assert client_log.split(b' ', 1)[-1].strip() == b'init commit'

    # Testing more commit and branch methods
    assert server_repo.commit_hash_title('master') == [sha.encode(),
                                                       b'init commit']
    assert server_repo.get_branch_sha(b'master') == sha



# Copyright 2019-2020 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import absolute_import

import pytest
from heptapod.testhelpers import RepoWrapper
from mercurial import (
    encoding,
    pycompat,
)

from .utils import common_config

parametrize = pytest.mark.parametrize


def make_repo(tmpdir, config):
    full_config = common_config()
    full_config.update(config)
    wrapper = RepoWrapper.init(tmpdir.join('repo'), config=full_config)
    return wrapper.repo.ui


def test_multiple_heads_allow(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(heptapod={'allow-multiple-heads': 'yes'},
                               experimental={'single-head-per-branch': 'no'},
                               ))
    assert not ui.configbool(b'experimental', b'single-head-per-branch')


def test_bookmarks_allow(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(heptapod={'allow-bookmarks': 'yes'},
                               ))
    assert not ui.configbool(b'experimental', b'single-head-per-branch')
    assert ui.configbool(b'experimental',
                         b'hg-git.bookmarks-on-named-branches')


def test_auto_publish_nothing(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(
                       heptapod={'auto-publish': 'nothing'},
                       experimental={'topic.publish-bare-branch': 'yes'},
                   ))
    assert not ui.configbool(b'experimental', b'topic.publish-bare-branch')


def test_auto_publish_all(tmpdir):
    ui = make_repo(tmpdir,
                   config=dict(
                       heptapod={'auto-publish': 'all'},
                       phases={'publish': 'no'},
                   ))
    assert ui.configbool(b'phases', b'publish')


@parametrize('key',
             [('experimental', 'single-head-per-branch'),
              ('experimental', 'hg-git.bookmarks-on-named-branches'),
              ('experimental', 'topic.publish-bare-branch'),
              ('phases', 'publish'),
              ])
@parametrize('value', [True, False])
def test_lower_level_config_untouched(tmpdir, key, value):
    """This case is the default config in Heptapod normal conditions."""
    section, item = key
    ui = make_repo(tmpdir, config={section: {item: value}})
    assert ui.configbool(pycompat.sysbytes(section),
                         pycompat.sysbytes(item)) == value


@parametrize('env_value,no_git_value,expected,expected_no_git',
             ((None, None, False, False),
              (None, 'yes', False, False),
              (None, 'no', False, False),
              ('no', 'yes', False, False),
              ('no', 'no', False, False),
              ('yes', None, True, False),
              ('yes', 'no', True, False),
              ('yes', 'yes', True, True),
              ))
def test_native_from_env(env_value, no_git_value, expected, expected_no_git,
                         tmpdir, monkeypatch):
    env = dict(encoding.environ)
    env_key = b'HEPTAPOD_HG_NATIVE'
    no_git_key = b'HEPTAPOD_NO_GIT'
    if env_value is not None:
        env_value = env_value.encode()
        env[env_key] = env_value
    if no_git_value is not None:
        no_git_value = no_git_value.encode()
        env[no_git_key] = no_git_value

    monkeypatch.setattr(encoding, 'environ', env)

    ui = make_repo(tmpdir, {})

    # confirm that test patching worked
    assert ui.environ.get(env_key) == env_value
    # actual test assertions
    assert ui.configbool(b'heptapod', b'native') is expected
    assert ui.configbool(b'heptapod', b'no-git') is expected_no_git
